@@ -1,121 +1,162 @@
-"""
-@author: The Alchemist
-@url: https://sp3rtah.github.io/sp3rtah
-@license: MIT
-"""
 import sys
 import json
+import time
 import loguru
 import asyncio
 import aiohttp
-import requests as _requests
-from queue import Queue, Empty
-from threading import Event
-from typing import Dict, List
-from contextlib import contextmanager
-from typing import Callable
+import requests
 import threading
-
+from typing import Callable
+from typing import Dict, Any
+from queue import Queue, Empty
+from contextlib import contextmanager
+from pyngrok import ngrok, conf as ngrokconf
+from concurrent.futures import ThreadPoolExecutor, CancelledError
 #
-from . import *
-from autogram.hooker import MyServer
+from autogram import *
+from autogram.webhook import MyServer
 from bottle import request, response, post, run
 
 
 class Autogram:
     api_url = 'https://api.telegram.org/'
-    config = default_config
-    media_quality = "high"
 
-    @classmethod
-    def getMediaQuality(cls):
-        if (qlty := Autogram.media_quality) == 'high':
+    def __init__(self, config: Dict):
+        self.config = config
+        self._initialize_()
+
+    def _initialize_(self):
+        self.webhook = None
+        self.host = '0.0.0.0'
+        self.update_offset = 0
+        self.httpRoutines = list()
+        self.httpRequests = Queue()
+        self.worker_threads = {
+            'io-tasks': list(),
+            'common-tasks': list(),
+            'high-priority': list(),
+        }
+        self.failing_endpoints = list()
+        self.terminate = threading.Event()
+        self.port = self.config['tcp-port']
+        self.control_lock = threading.Lock()
+        self.executor = ThreadPoolExecutor()
+        self.timeout = aiohttp.ClientTimeout(self.config['tcp-timeout'])
+        self.base_url = f"{Autogram.api_url}bot{self.config['telegram-token']}"
+        # 
+        logger_format = (
+            "<green>{time:DD/MM/YYYY HH:mm:ss}</green> | "
+            "<level>{level: <8}</level>|"
+            "<cyan>{line}</cyan>:<cyan>{name}</cyan>:<cyan>{function}</cyan> | "
+            "<level>{message}</level>"
+        )
+        loguru.logger.remove()
+        #
+        lvl = 'DEBUG'
+        if env := self.config.get('env'):
+            if env == 1:
+                lvl = self.config.get('log-level') or 'DEBUG'
+        #
+        loguru.logger.add(sys.stderr, format=logger_format, level=lvl)
+        self.logger = loguru.logger
+        self.control_lock.acquire()
+        self.logger.debug('Logs initialized successfully')
+
+    def mediaQuality(self):
+        if (qlty := self.config.get("media-quality") or 'high') == 'high':
             return 2
         elif qlty == 'medium':
             return 1
         return 0
 
-    def __init__(self, config: Dict):
-        """
-        class Autogram:
-        """
-        self.config = config
-        self.config['telegram_token'] = config['telegram_token']
-        self._initialize_()
-
-    def _initialize_(self):
-        self.timeout = aiohttp.ClientTimeout(self.config['tcp_timeout'])
-        self.base_url = f"{Autogram.api_url}bot{self.config['telegram_token']}"
-        # 
-        self.admin = None
-        self.deputy_admin = None
-        self.routines = set()
-        # 
-        self.port = 4004
-        self.webhook = None
-        self.host = '0.0.0.0'
-        self.update_offset = 0
-        self.requests = Queue()
-        self.terminate = Event()
-        # 
-        self.setup_logger()
-
-    def __repr__(self) -> str:
-        return str(vars(self))
-
     @loguru.logger.catch
     def send_online(self) -> threading.Thread:
         """Get this bot online in a separate daemon thread."""
-        if self.config['public_ip']:
-            hookPath = self.config['telegram_token'].split(":")[-1]
+        if public_ip := self.config['tcp-ip']:
+            hookPath = self.config['telegram-token'].split(":")[-1].lower()[:8]
             @post(f'/{hookPath}')
             def hookHandler():
                 self.updateRouter(request.json)
                 response.content_type = 'application/json'
                 return json.dumps({'ok': True})
             # 
-            def runServer(server: MyServer):
+            def runServer(server: Any):
                 run(server=server, quiet=True)
             # 
             server = MyServer(host=self.host,port=self.port)
-            svr_thread = threading.Thread(target=runServer,args=(server,))
-            svr_thread.name = 'Bottle Server'
-            svr_thread.daemon = True
-            svr_thread.start()
+            serv_thread = threading.Thread(target=runServer, args=(server,))
+            serv_thread.name = 'Autogram::Bottle'
+            serv_thread.daemon = True
+            serv_thread.start()
+            # check public ip availability
+            if public_ip == 'ngrok':
+                ngrok_config = ngrokconf.PyngrokConfig(
+                    ngrok_version='v3',
+                    ngrok_path='/snap/bin/ngrok',
+                    auth_token= self.config.get('ngrok-token')
+                )
+                ngrokconf.set_default(ngrok_config)
+                if log_ngrok := self.config.get('ngrok-logs'):
+                    ngrokconf.get_default().log_event_callback = self.logger.debug
+                try:
+                    self.ngrok_tunnel = ngrok.connect()
+                    if not log_ngrok:
+                        ngrok.get_ngrok_process().stop_monitor_thread()
+                    public_ip = self.ngrok_tunnel.public_url
+                except Exception as e:
+                    self.logger.critical(e)
+                    self.terminate.set()
             #
-            self.webhook = f"{self.config['public_ip']}/{hookPath}"
-            self.logger.info(f'Webhook set successfully...')
-        #
+            if not self.terminate.is_set():
+                self.webhook = f"{public_ip}/{hookPath}"
+                self.logger.debug(f'Webhook: {self.webhook}')
+        # wrap and start main_loop
         def launch():
+            if self.terminate.is_set():
+                return
             try:
                 if sys.platform != 'linux':
                     asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
                 asyncio.run(self.main_loop())
             except KeyboardInterrupt:
                 self.terminate.set()
-                _pending = self.requests.unfinished_tasks
-                if _pending:
-                    self.logger.info(f'[autogram]: {_pending} update tasks...')
-            except Exception:
-                raise
+            except Exception as e:
+                loguru.logger.exception(e)
+            finally:
+                self.terminate.set()
+        #
         worker = threading.Thread(target=launch)
         worker.name = 'Autogram'
         worker.daemon = True
         worker.start()
         return worker
 
-    def setup_logger(self):
-        """prepare logger and log levels"""
-        self.logger = loguru.logger
-
-    def getMe(self, me: Dict):
-        """receive and parse getMe request."""
-        self.logger.info('*** connected... ***')
-        for k,v in me.items():
-            setattr(self, k, v)
-
     @loguru.logger.catch()
-    def updateRouter(self, res: Dict|List):
+    async def main_loop(self):
+        """Main control loop"""
+        processor = asyncio.create_task(self.aioWebRequest())
+        self.httpRequests.put((f'{self.base_url}/getMe',None,self.getMe))
+        if self.webhook:
+            url = f'{self.base_url}/setWebhook'
+            self.httpRequests.put((url,{
+                'params': {
+                    'url': self.webhook
+                }
+            }, None))
+        else:   # delete webhook
+            def check_webhook(info: dict):
+                if info['url']:
+                    self.deleteWebhook()
+            self.getWebhookInfo(check_webhook)
+        #
+        await asyncio.sleep(0)    # allow getMe to run
+        await processor
+        #
+        self.terminate.set()
+        self.logger.info('Autogram terminated.')
+
+    @loguru.logger.catch
+    def updateRouter(self, res: Any):
         """receive and route updates"""
         def handle(update: Dict):
             parser = None
@@ -154,13 +195,13 @@ class Autogram:
             # 
             if not parser:
                 return
-            else:
-                if not self.admin and parser.name != 'message':
-                    return
-                parser.autogram = self
-                parse_thread = threading.Thread(target=parser,args=(payload,))
-                parse_thread.daemon = True
-                parse_thread.start()
+            # todo: implement all update types then allow through
+            if parser.name != 'message':
+                self.logger.critical(f"Unimplemented: {parser.name}")
+                return
+            #
+            parser.autogram = self
+            self.toThread(parser,payload)
         # 
         if type(res) == list:
             for update in res:
@@ -169,32 +210,45 @@ class Autogram:
                 handle(update)
             return
         handle(res)
+        return
 
-    async def main_loop(self):
-        """Main control loop"""
-        self.logger.debug('Main loop started...')
-        processor = asyncio.create_task(self.aioWebRequest())
-        self.requests.put((f'{self.base_url}/getMe',None,self.getMe))
-        if self.webhook:
-            url = f'{self.base_url}/setWebhook'
-            self.requests.put((url,{
-                'params': {
-                    'url': self.webhook
-                }
-            }, None))
-        else:   # delete webhook
-            def check_webhook(info: dict):
-                if info['url']:
-                    self.deleteWebhook()
-            self.getWebhookInfo(check_webhook)
-        ##
-        await asyncio.sleep(0)    # allow getMe to run
-        await processor
-
-        ## done processing. terminate
-        self.terminate.set()
-        await asyncio.gather(*self.routines)
-        self.logger.debug('Main loop terminated.')
+    @loguru.logger.catch()
+    def toThread(self, *args, callback: Callable = None, errHandler: Callable = None, priority : str = ''):
+        if self.terminate.is_set():
+            self.logger.debug(f"[{callable.__name__}] : Skipping execution on termination")
+            return
+        #
+        to_remove = list()
+        for key in self.worker_threads:
+            for task in self.worker_threads[key]:
+                tsk, cb, errh = task
+                if not tsk.done():
+                    continue
+                # fet results or exceptions
+                try:
+                    result = tsk.result()
+                    if cb:
+                        cb(result)
+                except Exception as e:
+                    if errh:
+                        errh(e)
+                    else:
+                        self.logger.exception(e)
+                to_remove.append((key, self.worker_threads[key].index(task)))
+        #
+        for item in to_remove:
+            key, index = item
+            self.worker_threads[key].pop(index)
+            continue
+        # append workers to priority groups
+        result = False
+        if not priority or (result := (priority not in list(self.worker_threads.keys()))):
+            if result:
+                self.logger.debug(f"Undefined task priority: {priority}")
+            priority = 'common-tasks'
+        task = self.executor.submit(*args)
+        self.worker_threads[priority].append((task, callback, errHandler))
+        return task
 
     @contextmanager
     def get_request(self):
@@ -205,9 +259,9 @@ class Autogram:
             return
         elif self.webhook:
             if not self.terminate.is_set():
-                if self.requests.empty():
+                if self.httpRequests.empty():
                     try:
-                        yield self.requests.get(timeout=3)
+                        yield self.httpRequests.get(timeout=3)
                     except Empty as e:
                         yield None
                     except Exception as e:
@@ -215,12 +269,12 @@ class Autogram:
                         self.terminate.set()
                         yield None
                 else:
-                    yield self.requests.get(block=False)
+                    yield self.httpRequests.get(block=False)
             else:
                 yield None
             return
-        if not self.requests.empty():
-            yield self.requests.get(block=False)
+        if not self.httpRequests.empty():
+            yield self.httpRequests.get(block=False)
             return
         yield None
 
@@ -228,20 +282,49 @@ class Autogram:
     async def aioWebRequest(self):
         """Make asynchronous requests to the Telegram API"""
         async with aiohttp.ClientSession(timeout=self.timeout) as session:
-            failed_requests = 0
             self.failed = None
-
-            ## waiting loop
+            #
+            @loguru.logger.catch
+            async def httpHandler():
+                while not self.terminate.is_set():
+                    if not self.httpRoutines:
+                        await asyncio.sleep(0)
+                        continue
+                    #
+                    for item in self.httpRoutines:
+                        incoming, outgoing = item
+                        #
+                        resp, payload = incoming
+                        _, _, callback = outgoing
+                        endpoint = link.split('/')[-1]
+                        if resp.ok:
+                            if endpoint in self.failing_endpoints:
+                                self.failing_endpoints.remove(endpoint)
+                            self.httpRoutines.remove(item)
+                            #
+                            if payload and callback:
+                                self.toThread(callback, payload)
+                                # if it was a getMe request, wait for it to finish
+                                if not self.control_lock.locked():
+                                    self.control_lock.acquire()
+                            continue
+                        elif endpoint not in self.failing_endpoints:
+                            if payload:
+                                self.logger.critical(f"[{endpoint}] HTTP{resp.status} : {endpoint} : {payload}")
+                            else:
+                                self.logger.critical(f"[{endpoint}] HTTP{resp.status} : {outgoing}")
+                        # ignore repeated output
+                        self.failing_endpoints.append(endpoint)
+                        continue
+                    # clear handled routines
+                    self.httpRoutines.clear()
+            #
+            # start httpHandler
+            asyncio.create_task(httpHandler())
             while not self.terminate.is_set():
-                if self.config['max_retries']:
-                    if failed_requests > self.config['max_retries']:
-                        self.logger.critical(f'Maximum retries reached: {failed_requests}. Shutting down...')
-                        self.terminate.set()
-                        break
-                # 
                 with self.get_request() as request:
                     if not request:
-                        ## no webhook? create getJob task
+                        ## no webhook? convert to getJob task
                         if not self.webhook:
                             params = {
                                 'params': {
@@ -252,7 +335,7 @@ class Autogram:
                             request = (url,params,self.updateRouter)
                         else:
                             continue
-                    link, kw, callback = request
+                    link, kw, _ = request
                     kw = kw or dict()
                     defaults = {
                         'params': {
@@ -266,62 +349,116 @@ class Autogram:
                     else:
                         kw['params'] |= defaults['params']
                     ##
-                    error_detected = None
                     try:
-                        self.logger.debug(f'get: {link.split("/")[-1]}')
+                        error_detected = None
                         async with session.get(link,**kw) as resp:
+                            data = None
                             if resp.ok:
                                 data = await resp.json()
-                                if callback:
-                                    payload = data['result']
-                                    if payload:
-                                        callback(payload)
-                                if failed_requests > 0:
-                                   failed_requests -= 1
-                                   self.failed = None
+                                self.logger.debug(f'GET [ok] /{link.split("/")[-1]}')
                             else:
-                                self.logger.critical(f"{resp.status},{await resp.json()} (endpoint:{link.split('/')[-1]}: kw: {kw}: callback: {callback})")
-                                self.terminate.set()
-                                break
+                                self.logger.debug(f'GET [err] /{link.split("/")[-1]} : {kw}')
+                            self.httpRoutines.append(((resp, data), request))
                     except KeyboardInterrupt:
                         self.terminate.set()
-                    except aiohttp.client_exceptions.ClientConnectorError as e:
-                        self.terminate.set()
-                        self.logger.exception(e)
+                        self.logger.critical('Termination with ^C')
+                    except aiohttp.ClientConnectorError as e:
                         error_detected = e
-                    except aiohttp.client_exceptions.ClientOSError as e:
+                        self.logger.critical('ClientConnectorError')
+                    except aiohttp.ClientOSError as e:
                         error_detected = e
-                    except asyncio.exceptions.TimeoutError as e:
+                        self.logger.critical('ClientOSError')
+                    except asyncio.TimeoutError as e:
                         error_detected = e
+                        self.logger.critical('TimeoutError')
                     except Exception as e:
-                        self.terminate.set()
-                        self.logger.exception(e)
                         error_detected = e
+                        self.logger.critical(f'Unknown Error: {e}')
                     finally:
                         if error_detected:
-                            failed_requests += 1
-                            print(error_detected)
-                            self.failed = (link,kw,callback)
-
+                            self.failed = request
+                            self.logger.exception(error_detected)
+                #
+                await asyncio.sleep(0)
 
     @loguru.logger.catch()
     def webRequest(self, url: str, params={}, files=None):
+        res = None
         params = params or {}
         # send request
         try:
             if files:
-                res = _requests.get(url,params=params,files=files)
+                res = requests.get(url,params=params,files=files)
             else:
-                res = _requests.get(url,params=params)
+                res = requests.get(url,params=params)
+            #
+            if res.ok:
+                return True, json.loads(res.text)['result']
         except Exception as e:
-            self.logger.critical(f'webRequest failed: {url.split("/")[-1]}')
             self.logger.exception(e)
-            return
-        ##
-        if not res.ok:
+        finally:
             return False, res, url
-        return True, json.loads(res.text)['result']
 
+    def shutdown(self, callback: Callable = None):
+        """callback: your exit function that takes `msg : str`"""
+        self.terminate.set()
+        exit_msg = list()
+        critical_priority = ['io-threads', 'high-priority']
+        for key in critical_priority:
+            for task in self.worker_threads[key]:
+                tsk, cb, errh = task
+                if not tsk.done() and errh:
+                    if tsk.cancel():
+                        errh(CancelledError(f'Thread leading to {cb.__name__} was canceled'))
+                    else:
+                        errh(CancelledError(f'Thread leading to {cb.__name__} rejected termination.'))
+                elif not tsk.done():
+                    exit_msg.append(f"thread to: {cb.__name__} ~ busy")
+        #
+        msg = '\n'.join(exit_msg)
+        if callback:
+            callback(msg)
+        else:
+            self.logger.debug(msg)
+
+    #***** start API calls *****#
+    def getMe(self, me: Dict):
+        """receive and parse getMe request."""
+        self.logger.info('*** connected... ***')
+        for k,v in me.items():
+            setattr(self, k, v)
+        # 
+        self.control_lock.release()
+
+    @loguru.logger.catch()
+    def downloadFile(self, file_path: str):
+        url = f"https://api.telegram.org/file/bot{self.config['telegram-token']}/{file_path}"
+        res = requests.get(url)
+        if res.ok:
+            return res.content
+        else:
+            self.logger.critical(f'file: [{file_path} -> Download failed: {res.status_code}')
+
+    @loguru.logger.catch()
+    def getFile(self, file_id: str):
+        url = f'{self.base_url}/getFile'
+        return self.webRequest(url, params={'file_id' : file_id})
+
+    @loguru.logger.catch()
+    def getChat(self, chat_id: int, handler: Callable):
+        url = f'{self.base_url}/getChat'
+        self.httpRequests.put((url, {
+            'params': {
+                'chat_id': chat_id
+            }
+        }, handler))
+
+    @loguru.logger.catch()
+    def getWebhookInfo(self, handler: Callable):
+        url = f'{self.base_url}/getWebhookInfo'
+        self.httpRequests.put((url,None,handler))
+
+    @loguru.logger.catch()
     def sendChatAction(self, chat_id: int, action: str):
         params = {
             'chat_id': chat_id,
@@ -329,57 +466,35 @@ class Autogram:
         }
         return self.webRequest(f'{self.base_url}/sendChatAction', params=params)
 
-    def getFile(self, file_id: str):
-        url = f'{self.base_url}/getFile'
-        params = { 'file_id': file_id }
-        return self.webRequest(url, params=params)
-
-    def getChat(self, chat_id: int, handler: Callable):
-        url = f'{self.base_url}/getChat'
-        self.requests.put((url, {
-            'params': {
-                'chat_id': chat_id
-            }
-        }, handler))
-
-    def getWebhookInfo(self, handler: Callable):
-        url = f'{self.base_url}/getWebhookInfo'
-        self.requests.put((url,None,handler))
-
     @loguru.logger.catch()
-    def downloadFile(self, file_path: str):
-        url = f"https://api.telegram.org/file/bot{self.config['telegram_token']}/{file_path}"
-        res = _requests.get(url)
-        if res.ok:
-            return res.content
-        else:
-            self.logger.critical(f'file: [{file_path} -> Download failed: {res.status_code}')
-
-    @loguru.logger.catch()
-    def sendMessage(self, chat_id: int, text: str, params={}):
+    def sendMessage(self, chat_id: int, text: str, **kwargs):
         url = f'{self.base_url}/sendMessage'
-        self.requests.put((url, {
+        self.httpRequests.put((url,{
             'params': {
                 'chat_id': chat_id,
-                'text': text
-            }|params
+                'text': text,
+            } | kwargs
+        } , None))
+
+    @loguru.logger.catch()
+    def deleteMessage(self, chat_id: int, msg_id: int):
+        url = f'{self.base_url}/deleteMessage'
+        self.httpRequests.put((url,{
+            'params': {
+                'chat_id': chat_id,
+                'message_id': msg_id
+            }
         }, None))
 
     @loguru.logger.catch()
-    def forwardMessage(self, chat_id: int, from_chat_id: int, msg_id: int):
-        url = f'{self.base_url}/forwardMessage'
-        self.requests.put((url,{
-            'params': {
-                'chat_id': chat_id,
-                'from_chat_id': from_chat_id,
-                'message_id': msg_id
-            }
-        },None))
+    def deleteWebhook(self):
+        url = f'{self.base_url}/deleteWebhook'
+        return self.webRequest(url)
 
     @loguru.logger.catch()
     def editMessageText(self, chat_id: int, msg_id: int, text: str, params={}):
         url = f'{self.base_url}/editMessageText'
-        self.requests.put((url,{
+        self.httpRequests.put((url,{
             'params': {
                 'text':text,
                 'chat_id': chat_id,
@@ -390,7 +505,7 @@ class Autogram:
     @loguru.logger.catch()
     def editMessageCaption(self, chat_id: int, msg_id: int, capt: str, params={}):
         url = f'{self.base_url}/editMessageCaption'
-        self.requests.put((url, {
+        self.httpRequests.put((url, {
             'params': {
                 'chat_id': chat_id,
                 'message_id': msg_id,
@@ -401,7 +516,7 @@ class Autogram:
     @loguru.logger.catch()
     def editMessageReplyMarkup(self, chat_id: int, msg_id: int, markup: str, params={}):
         url = f'{self.base_url}/editMessageReplyMarkup'
-        self.requests.put((url,{
+        self.httpRequests.put((url,{
             'params': {
                 'chat_id':chat_id,
                 'message_id':msg_id,
@@ -410,19 +525,17 @@ class Autogram:
         }, None))
 
     @loguru.logger.catch()
-    def deleteMessage(self, chat_id: int, msg_id: int):
-        url = f'{self.base_url}/deleteMessage'
-        self.requests.put((url,{
+    def forwardMessage(self, chat_id: int, from_chat_id: int, msg_id: int):
+        url = f'{self.base_url}/forwardMessage'
+        self.httpRequests.put((url,{
             'params': {
                 'chat_id': chat_id,
+                'from_chat_id': from_chat_id,
                 'message_id': msg_id
             }
-        }, None))
-
-    def deleteWebhook(self):
-        url = f'{self.base_url}/deleteWebhook'
-        return self.webRequest(url)
+        },None))
 
+    @loguru.logger.catch()
     def answerCallbackQuery(self, query_id,text:str= None, params : dict= None):
         url = f'{self.base_url}/answerCallbackQuery'
         params.update({
@@ -431,6 +544,7 @@ class Autogram:
         })
         return self.webRequest(url, params)
 
+    @loguru.logger.catch()
     def sendPhoto(self,chat_id: int, photo_bytes: bytes, caption: str= None, params: dict= None):
         params = params or {}
         url = f'{self.base_url}/sendPhoto'
@@ -441,6 +555,7 @@ class Autogram:
         self.sendChatAction(chat_id,chat_actions.photo)
         return self.webRequest(url,params=params,files={'photo':photo_bytes})
 
+    @loguru.logger.catch()
     def sendAudio(self,chat_id: int,audio_bytes: bytes, caption: str= None, params: dict= None):
         params = params or {}
         url = f'{self.base_url}/sendAudio'
@@ -451,6 +566,7 @@ class Autogram:
         self.sendChatAction(chat_id,chat_actions.audio)
         return self.webRequest(url,params,files={'audio':audio_bytes})
 
+    @loguru.logger.catch()
     def sendDocument(self,chat_id: int ,document_bytes: bytes, caption: str= None, params: dict= None):
         params = params or {}
         url = f'{self.base_url}/sendDocument'
@@ -461,6 +577,7 @@ class Autogram:
         self.sendChatAction(chat_id,chat_actions.document)
         return self.webRequest(url,params,files={'document':document_bytes})
 
+    @loguru.logger.catch()
     def sendVideo(self,chat_id: int ,video_bytes: bytes, caption: str = None, params: dict= None ):
         params = params or {}
         url = f'{self.base_url}/sendVideo'
@@ -471,6 +588,7 @@ class Autogram:
         self.sendChatAction(chat_id,chat_actions.video)
         return self.webRequest(url,params,files={'video':video_bytes})
 
+    @loguru.logger.catch()
     def forceReply(self, params: dict= None):
         params = params or {}
         markup = {
@@ -478,6 +596,7 @@ class Autogram:
         }|params
         return json.dumps(markup)
 
+    @loguru.logger.catch()
     def getKeyboardMarkup(self, keys: list, params: dict= None):
         params = params or {}
         markup = {
@@ -485,6 +604,7 @@ class Autogram:
         }|params
         return json.dumps(markup)
 
+    @loguru.logger.catch()
     def getInlineKeyboardMarkup(self, keys: list, params: dict= None):
         params = params or {}
         markup = {
@@ -492,13 +612,19 @@ class Autogram:
         }|params
         return json.dumps(markup)
 
+    @loguru.logger.catch()
     def parseFilters(self, filters: dict= None):
         keys = list(filters.keys())
         return json.dumps(keys)
 
+    @loguru.logger.catch()
     def removeKeyboard(self, params: dict= None):
         params = params or {}
         markup = {
             'remove_keyboard': True,
         }|params
         return json.dumps(markup)
+
+    def __repr__(self) -> str:
+        return f"Autogram({self.config})"
+
